/*
 * unrpath - a rpath remover for ELF objects
 *
 * Copyright (C) 2006 Philippe Troin <phil@fifi.org>.
 *
 * Sponsored by Splunk http://www.splunk.com/
 *
 * "It was hard to write, it should be hard to understand."
 *
 * Clue: read some ELF docs on the net.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#if HAVE_LIBELF_LIBELF_H
#  include <libelf/libelf.h>
#else /* ! HAVE_LIBELF_LIBELF_H */
#  if HAVE_LIBELF_H
#    include <libelf.h>
#  else /* ! HAVE_LIBELF_H */
#    error libelf.h not found
#  endif /* HAVE_LIBELF_H */
#endif /* HAVE_LIBELF_LIBELF_H */

#if HAVE_LIBELF_GELF_H
#  include <libelf/gelf.h>
#else /* ! HAVE_LIBELF_GELF_H */
#  if HAVE_GELF_H
#    include <gelf.h>
#  else /* ! HAVE_GELF_H */
#    error gelf.h not found
#  endif /* HAVE_GELF_H */
#endif /* HAVE_LIBELF_GELF_H */


enum {
  XIT_OK,
  XIT_ARGS,
  XIT_WARNINGS,
  XIT_LIBELF,
};

int
main(int argc, char*argv[])
{
  const char*	progname;
  int		prognamelen;
  int		fileindex;
  int		nnonelf  = 0;
  int		nfailed  = 0;
  int		nupdated = 0;
  /**/

  progname = strrchr(argv[0], '/');
  if (progname)
    ++progname;
  else
    progname = argv[0];
  prognamelen = strlen(progname);

  if (argc <= 1
      || (argc == 2 && ! strcmp(argv[1], "--help")))
    {
      fprintf(argc == 2 ? stdout : stderr,
	      "usage: %s <elf-object>...\n", progname);
      exit(argc == 2 ? XIT_OK : XIT_ARGS);
    }

  if (argc == 2 && ! strcmp(argv[1], "--version"))
    {
      fprintf(stdout, "%s\nWritten by Philippe Troin <phil@fifi.org>.\n"
	      "Sponsored by Splunk ( http://www.splunk.com/ ).\n\n"
	      "Copyright (C) 2006 Philippe Troin <phil@fifi.org>\n"
	      "This is free software; see the source for copying conditions.  "
	      "There is NO\nwarranty; not even for MERCHANTABILITY or "
	      "FITNESS FOR A PARTICULAR PURPOSE.\n",
	      PACKAGE_STRING);
      exit(XIT_OK);
    }

  if (elf_version(EV_CURRENT) == EV_NONE)
    {
      fprintf(stderr, "%s: ELF version out of date: %s\n",
	      progname, elf_errmsg(elf_errno()));
      exit(XIT_LIBELF);
    }

  for (fileindex = 1; fileindex < argc; ++fileindex)
    {
      const char*	filename   = argv[fileindex];
      int		fd;
      Elf*		elf;
      GElf_Ehdr		gelfhdr_s;
      GElf_Ehdr*	gelfhdr;
      GElf_Dyn*         gelfdyn;
      GElf_Dyn          gelfdyn_s;
      Elf_Scn*		section;
      GElf_Shdr*	sectionhdr;
      GElf_Shdr		sectionhdr_s;
      Elf_Data*		data;
      int		oksofar    = 0;
      off_t		offset	   = 0;
      off_t		dynentsize = 0;
      unsigned          movecount  = 0;
      /**/

      fd = open(filename, O_RDWR);
      if (fd == -1)
	{
	  fprintf(stderr, "%s: %s: cannot open: %s\n",
		  progname, filename, strerror(errno));
	  goto L_next;
	}

    elf = elf_begin(fd, ELF_C_READ, 0);
    if (!elf)
      {
	fprintf(stderr, "%s: %s: cannot read ELF object: %s\n",
		progname, filename, elf_errmsg(elf_errno()));
	goto L_close;
      }

    if (elf_kind(elf) != ELF_K_ELF)
      {
	fprintf(stderr, "%s: %s: not an ELF object\n",
		progname, filename);
	++nnonelf;
	oksofar = 1;
	goto L_elf_close;
      }

    gelfhdr = gelf_getehdr(elf, &gelfhdr_s);
    if (!gelfhdr)
      {
	fprintf(stderr, "%s: %s: cannot get ELF header: %s\n",
		progname, filename, elf_errmsg(elf_errno()));
	goto L_elf_close;
      }

    switch(gelfhdr->e_type)
      {
      case ET_EXEC:
      case ET_DYN:
	break;
      case ET_REL:
	fprintf(stderr,
		"%s: %s: not making sense on relocatable (.o) ELF files\n",
		progname, filename);
	goto L_elf_close;
      case ET_CORE:
	fprintf(stderr,
		"%s: %s: not making sense on ELF core files\n",
		progname, filename);
	goto L_elf_close;
      case ET_NUM:
	fprintf(stderr,
		"%s: %s: not making sense on ELF NUM files\n",
		progname, filename);
	goto L_elf_close;
      case ET_LOOS:
	fprintf(stderr,
		"%s: %s: not making sense on ELF LOOS files\n",
		progname, filename);
	goto L_elf_close;
      case ET_HIOS:
	fprintf(stderr,
		"%s: %s: not making sense on ELF HIOS files\n",
		progname, filename);
	goto L_elf_close;
      case ET_LOPROC:
	fprintf(stderr,
		"%s: %s: not making sense on ELF LOPROC files\n",
		progname, filename);
	goto L_elf_close;
      case ET_HIPROC:
	fprintf(stderr,
		"%s: %s: not making sense on ELF HIPROC files\n",
		progname, filename);
	goto L_elf_close;
      default:
	fprintf(stderr,
		"%s: %s: not making sense on unknown ELF type %d\n",
		progname, filename, (int)gelfhdr->e_type);
	goto L_elf_close;
      }

    for (section=0; (section=elf_nextscn(elf, section)); /*empty*/)
      {
	sectionhdr = gelf_getshdr(section, &sectionhdr_s);
	if (!sectionhdr)
	  {
	    fprintf(stderr,
		    "%s: %s: cannot retrieve section header: %s\n",
		    progname, filename, elf_errmsg(elf_errno()));
	    goto L_elf_close;
	  }
	if (sectionhdr->sh_type == SHT_DYNAMIC)
	  for (data = 0; (data = elf_getdata(section, data)) ; /*empty*/)
	    if (data->d_type == ELF_T_DYN)
	      {
		if (data->d_size % sectionhdr->sh_entsize)
		  {
		    fprintf(stderr,
			    "%s: %s: partial entry in dynamic section!",
			    progname, filename);
		    goto L_elf_close;
		  }
		else
		  {
		    int count = data->d_size / sectionhdr->sh_entsize;
		    int i;

		    for (i = 0; i < count; ++i)
		      {
			if (!(gelfdyn = gelf_getdyn(data, i, &gelfdyn_s)))
			  {
			    fprintf(stderr,
				    "%s: %s: cannot retrieve entry %d/%d "
				    "in dynamic section\n",
				    progname, filename, i, count);
			    goto L_elf_close;
			  }

			switch (gelfdyn->d_tag)
			  {
			  case DT_RPATH:
			    if (dynentsize)
			      {
				fprintf(stderr,
					"%s: %s: found two DT_RPATH entries\n",
					progname, filename);
				goto L_elf_close;
			      }
			    dynentsize = sectionhdr->sh_entsize;
			    offset	 = sectionhdr->sh_offset
			      + i*sectionhdr->sh_entsize;
			    movecount  = count-i-1;
			    break;
			  }
		      }
		    // Last entry should be DT_NULL
		    if (gelfdyn->d_tag != DT_NULL)
		      {
			fprintf(stderr,
				"%s: %s: last entry on dynamic section was "
				"not DT_NULL but %lld\n",
				progname, filename, (long long)gelfdyn->d_tag);
			goto L_elf_close;
		      }
		  }
	      }
	    else
	      {
		fprintf(stderr,
			"%s: %s: found non ELF_T_DYN data in SHT_DYNAMIC "
			"section",
			progname, filename);
		goto L_elf_close;
	      }
      }

    if (dynentsize)
      {
	char	buf[movecount*dynentsize];
	ssize_t nread;
	ssize_t nwritten;

	if (lseek(fd, offset+dynentsize, SEEK_SET) == (off_t)-1)
	  {
	    fprintf(stderr, "%s: %s: cannot seek to offset %lld: %s\n",
		    progname, filename, (long long)(offset+dynentsize),
		    strerror(errno));
	    goto L_elf_close;
	  }

	nread = read(fd, buf, movecount*dynentsize);
	if (nread == -1)
	  {
	    fprintf(stderr,
		    "%s: %s: cannot read %lld bytes at offset %lld: %s\n",
		    progname, filename, (long long)(movecount*dynentsize),
		    (long long)(offset+dynentsize), strerror(errno));
	    goto L_elf_close;
	  }
	if (nread != movecount*dynentsize)
	  {
	    fprintf(stderr,
		    "%s: %s: short read of %lld bytes (wanted %lld bytes) "
		    "at offset %lld\n",
		    progname, filename, (long long)nread,
		    (long long)(movecount*dynentsize),
		    (long long)(offset+dynentsize));
	    goto L_elf_close;
	  }

	if (lseek(fd, offset, SEEK_SET) == (off_t)-1)
	  {
	    fprintf(stderr, "%s: %s: cannot seek to offset %lld: %s\n",
		    progname, filename, (long long)offset,
		    strerror(errno));
	    goto L_elf_close;
	  }

	nwritten = write(fd, buf, nread);
	if (nwritten == -1)
	  {
	    fprintf(stderr,
		    "%s: %s: cannot write %lld bytes at offset %lld: %s\n",
		    progname, filename, (long long)nread, (long long)offset,
		    strerror(errno));
	    goto L_elf_close;
	  }
	if (nwritten != nread)
	  {
	    fprintf(stderr,
		    "%s: %s: short write of %lld bytes (wanted %lld bytes) "
		    "at offset %lld\n",
		    progname, filename, (long long)nwritten,
		    (long long)nread, (long long)offset);
	    goto L_elf_close;
	  }
	oksofar = 1;
	++nupdated;
      }
    else
      {
	oksofar=1;
	fprintf(stderr, "%s: %s: no DT_RPATH found\n",
		progname, filename);
      }

  L_elf_close:
    elf_end(elf);

  L_close:
    if (close(fd) == -1) {
      oksofar = 0;
      fprintf(stderr, "%s: %s: cannot close: %s\n",
	      progname, filename, strerror(errno));
    }

  L_next:
    if (!oksofar)
      ++nfailed;
  }

  fprintf(stderr,
	  "%s: summary: out of %d files examined:\n"
	  "%*s           %5d were not ELF files\n"
	  "%*s           %5d did not contain DT_RPATH tags\n"
	  "%*s           %5d had DT_RPATH removed\n"
	  "%*s           %5d had an error of some sort\n",
	  progname, argc-1,
	  prognamelen, "", nnonelf,
	  prognamelen, "", argc-1-nnonelf-nupdated-nfailed,
	  prognamelen, "", nupdated,
	  prognamelen, "", nfailed);

  if (nfailed || nnonelf)
    exit(XIT_WARNINGS);
  else
    exit(XIT_OK);
}
